import numpy as np
import pandas as pd
from typing import Optional

from .ode import OdeModel
from .vpop import generate_vpop_from_ranges
from .structural_model import StructuralOdeModel
from .nlme import NlmeModel


def simulate_dataset_from_ranges(
    ode_model: OdeModel,
    log_nb_individuals: int,
    param_ranges: dict[str, dict[str, float | bool]],
    initial_conditions: np.ndarray,
    protocol_design: Optional[pd.DataFrame],
    residual_error_variance: Optional[np.ndarray],
    error_model: Optional[str],  # "additive" or "proportional"
    time_steps: np.ndarray,
) -> pd.DataFrame:
    """Generate a simulated data set with an ODE model

    Simulates a dataset for training a surrogate model. Timesteps can be different for each output.
    The parameter space is explored with Sobol sequences.

    Args:
        log_nb_individuals (int): The number of simulated patients will be 2^this parameter
        param_ranges (list[dict]): For each parameter in the model, a dict describing the search space 'low': low bound, 'high': high bound, and 'log': True if the search space is log-scaled
        initial_conditions (array): set of initial conditions, one for each variable
        protocol_design (optional): a DataFrame with a `protocol_arm` column, and one column per parameter override
        residual_error_variance (np.array): A 1D array of residual error variances for each output.
        error_model (str): the type of error model ("additive" or "proportional").
        time_steps (np.array): an array with the time points
    Returns:
        pd.DataFrame: A DataFrame with columns 'id', parameter names, 'time', 'output_name', and 'value'.

    Notes:
        If a parameter appears both in the ranges and in the protocol design, the ranges take precedence.
    """

    # Validate input data
    params_to_explore = list(param_ranges.keys())

    if protocol_design is None:
        print("No protocol")
        params = params_to_explore
        params_in_protocol = []
        protocol_design_filt = pd.DataFrame({"protocol_arm": ["identity"]})
    else:
        params_in_protocol = protocol_design.drop(
            "protocol_arm", axis=1
        ).columns.tolist()
        # Find the paramaters that appear both in the ranges and the protocol
        overlap = set(params_to_explore) & set(params_in_protocol)
        if overlap != set():
            protocol_design_filt = protocol_design.drop(list(overlap), axis=1)
            print(
                f"Warning: ignoring entries {overlap} from the protocol design (already defined in the ranges)."
            )
        else:
            protocol_design_filt = protocol_design

        params = params_to_explore + params_in_protocol
    if set(params) != set(ode_model.param_names):
        raise ValueError(
            f"Under-defined system: missing {set(ode_model.param_names) - set(params)}"
        )
    # Generate the vpop using sobol sequences
    patients_df = generate_vpop_from_ranges(log_nb_individuals, param_ranges)

    # Add a choice of protocol arm for each patient
    protocol_arms = pd.DataFrame(protocol_design_filt["protocol_arm"].drop_duplicates())
    patients_df = patients_df.merge(protocol_arms, how="cross")
    # Add the outputs for each patient
    outputs = pd.DataFrame({"output_name": ode_model.variable_names})
    patients_df = patients_df.merge(outputs, how="cross")
    # Simulate the ODE model
    output_df = ode_model.run_trial(
        patients_df, initial_conditions, protocol_design_filt, time_steps
    )
    # Pivot to wide to add noise per model output
    wide_output = output_df.pivot_table(
        index=["id", *ode_model.param_names, "time", "protocol_arm"],
        columns="output_name",
        values="predicted_value",
    ).reset_index()

    if error_model is None:
        pass
    else:
        if residual_error_variance is None:
            raise ValueError("Undefined residual error variance.")
        else:
            # Add noise to the data
            noise = np.random.normal(
                np.zeros_like(residual_error_variance),
                np.sqrt(residual_error_variance),
                (wide_output.shape[0], ode_model.nb_outputs),
            )
            if error_model == "additive":
                wide_output[ode_model.variable_names] += noise
            elif error_model == "proportional":
                wide_output[ode_model.variable_names] += (
                    noise * wide_output[ode_model.variable_names]
                )
            else:
                raise ValueError(f"Incorrect error_model choice: {error_model}")
    # Pivot back to long format
    long_output = wide_output.melt(
        id_vars=[
            "id",
            "protocol_arm",
            "time",
            *ode_model.param_names,
        ],
        value_vars=ode_model.variable_names,
        var_name="output_name",
        value_name="value",
    )
    # Remove the protocol arm overrides from the data set, they described by the protocol_arm column now
    long_output = long_output.drop(params_in_protocol, axis=1)
    return long_output


def simulate_dataset_from_omega(
    ode_model: OdeModel,
    protocol_design: pd.DataFrame,
    time_steps: np.ndarray,
    init_conditions: np.ndarray,
    log_mi: dict[str, float],
    log_pdu: dict[str, dict[str, float]],
    error_model: str,
    res_var: list[float],
    covariate_map: dict[str, dict[str, dict[str, str | float]]],
    patient_covariates: pd.DataFrame,
) -> pd.DataFrame:
    """Generate synthetic data set using an ODE model and population distributions of parameters

    Args:
        ode_model (OdeModel): The equations to be simulated
        protocol_design (pd.DataFrame): _description_
        time_steps (np.ndarray): _description_
        init_conditions (np.ndarray): _description_
        log_mi (dict[str, float]): _description_
        log_pdu (dict[str, dict[str, float]]): _description_
        error_model (str): _description_
        res_var (list[float]): _description_
        covariate_map (dict[str, dict[str, dict[str, str  |  float]]]): _description_
        patient_covariates (pd.DataFrame): _description_

    Returns:
        pd.DataFrame: _description_
    """

    structural_model = StructuralOdeModel(ode_model, protocol_design, init_conditions)
    nlme_model = NlmeModel(
        structural_model,
        patient_covariates,
        log_mi,
        log_pdu,
        res_var,
        covariate_map,
        error_model,
        num_chains=1,
    )
    etas = nlme_model.sample_etas(nlme_model.num_chains, nlme_model.nb_patients)
    phi = nlme_model.etas_to_gaussian_params(etas)
    pdu = nlme_model.gaussian_to_physical_params(phi, nlme_model.log_MI)
    theta = nlme_model.assemble_individual_parameters(pdu).squeeze(0)
    vpop = pd.DataFrame(data=theta.cpu().numpy(), columns=nlme_model.descriptors)
    vpop["id"] = nlme_model.patients
    protocol_arms = patient_covariates[["id", "protocol_arm"]]
    vpop = vpop.merge(protocol_arms, on=["id"], how="left")
    vpop = vpop.merge(
        pd.DataFrame(data=nlme_model.outputs_names, columns=["output_name"]),
        how="cross",
    )
    time_df = pd.DataFrame(data=time_steps, columns=["time"])
    vpop = vpop.merge(time_df, how="cross")
    # add a dummy observation value
    vpop["value"] = 1.0
    nlme_model.add_observations(vpop)

    out_tensor, _ = nlme_model.predict_outputs_from_theta(theta)
    out_with_noise = nlme_model.add_residual_error(out_tensor)
    out_df = nlme_model.outputs_to_df(out_with_noise)
    out_df = out_df.rename(columns={"predicted_value": "value"})

    return out_df



import torch
import pandas as pd
import numpy as np
import uuid
from typing import Callable, Optional
import itertools

from .model.gp import GP
from .ode import OdeModel
from .utils import device


class StructuralModel:
    def __init__(
        self,
        parameter_names,
        output_names,
        protocol_arms,
        tasks,
        task_idx_to_output_idx,
        task_idx_to_protocol,
    ):
        """Initialize a structural model

        Args:
            parameter_names (list[str]): _description_
            output_names (list[str]): _description_
            protocol_arms (list[str]): _description_
            tasks (list[str]): _description_
            task_idx_to_output_idx (list[str]): _description_
            task_idx_to_protocol (list[str]): _description_
        """
        self.parameter_names: list[str] = parameter_names
        self.nb_parameters: int = len(self.parameter_names)
        self.output_names: list[str] = output_names
        self.nb_outputs: int = len(self.output_names)
        self.protocols: list[str] = protocol_arms
        self.nb_protocols: int = len(self.protocols)
        self.tasks: list[str] = tasks
        self.task_idx_to_output_idx: dict[int, int] = task_idx_to_output_idx
        self.task_idx_to_protocol: dict[int, str] = task_idx_to_protocol

    def simulate(
        self,
        X: torch.Tensor,
        prediction_index: tuple[torch.Tensor, torch.Tensor, torch.Tensor],
        chunks: list[int],
    ) -> tuple[torch.Tensor, torch.Tensor]:
        raise ValueError("Not implemented")


class StructuralGp(StructuralModel):
    def __init__(self, gp_model: GP):
        """Create a structural model from a GP

        Args:
            gp_model (GP): The trained GP
        """
        # list the GP parameters, except time, as it will be handled differently in the NLME model
        parameter_names = [p for p in gp_model.data.parameter_names if p != "time"]
        super().__init__(
            parameter_names,
            gp_model.data.output_names,
            gp_model.data.protocol_arms,
            gp_model.data.tasks,
            gp_model.data.task_idx_to_output_idx,
            gp_model.data.task_idx_to_protocol,
        )
        self.gp_model = gp_model
        self.training_ranges = {}
        training_samples = self.gp_model.data.full_df_raw[self.parameter_names]
        train_min = training_samples.min(axis=0)
        train_max = training_samples.max(axis=0)
        for param in self.parameter_names:
            self.training_ranges.update(
                {param: {"low": train_min[param], "high": train_max[param]}}
            )

    def simulate(
        self,
        X: torch.Tensor,
        prediction_index: tuple[torch.Tensor, torch.Tensor, torch.Tensor],
        chunks: list[int],
    ) -> tuple[torch.Tensor, torch.Tensor]:
        (num_chains, nb_patients, nb_timesteps, nb_params) = X.shape
        nb_obs_per_chain = prediction_index[0].shape[0]
        prediction_index_expanded = (
            torch.arange(num_chains).repeat_interleave(nb_obs_per_chain),
            prediction_index[0].repeat(num_chains),
            prediction_index[1].repeat(num_chains),
            prediction_index[2].repeat(num_chains),
        )
        # Simulate the GP
        X_vertical = X.view(-1, nb_params)
        out_cat, var_cat = self.gp_model.predict_wide_scaled(X_vertical)
        out_wide = out_cat.view(num_chains, nb_patients, nb_timesteps, -1)
        var_wide = var_cat.view(num_chains, nb_patients, nb_timesteps, -1)

        # Retrieve the necessary rows and columns to transform into a single column tensor
        y = out_wide[prediction_index_expanded].view(num_chains, nb_obs_per_chain)
        var = var_wide[prediction_index_expanded].view(num_chains, nb_obs_per_chain)
        return y, var


class StructuralOdeModel(StructuralModel):
    def __init__(
        self,
        ode_model: OdeModel,
        protocol_design: pd.DataFrame,
        init_conditions: np.ndarray,
    ):
        self.ode_model = ode_model
        protocol_arms = protocol_design["protocol_arm"].drop_duplicates().to_list()
        self.protocol_design = protocol_design
        output_names: list[str] = self.ode_model.variable_names
        tasks: list[str] = [
            output + "_" + protocol
            for protocol in protocol_arms
            for output in output_names
        ]
        # Map tasks to output names
        task_to_output = {
            output_name + "_" + protocol_arm: output_name
            for output_name in output_names
            for protocol_arm in protocol_arms
        }
        # Map task index to output index
        task_idx_to_output_idx = {
            tasks.index(k): output_names.index(v) for k, v in task_to_output.items()
        }
        # Map task to protocol arm
        task_to_protocol = {
            output_name + "_" + protocol_arm: protocol_arm
            for output_name in output_names
            for protocol_arm in protocol_arms
        }
        # Map task index to protocol arm
        task_idx_to_protocol = {tasks.index(k): v for k, v in task_to_protocol.items()}

        # list the structural model parameters: the protocol overrides are ignored
        self.protocol_overrides = self.protocol_design.drop(
            columns="protocol_arm"
        ).columns.to_list()
        parameter_names = list(
            set(self.ode_model.param_names) - set(self.protocol_overrides)
        )
        self.nb_protocol_overrides = len(self.protocol_overrides)

        super().__init__(
            parameter_names,
            output_names,
            protocol_arms,
            tasks,
            task_idx_to_output_idx,
            task_idx_to_protocol,
        )

        self.init_cond_df = pd.DataFrame(
            data=[init_conditions], columns=self.ode_model.initial_cond_names
        )

    def simulate(
        self,
        X: torch.Tensor,
        prediction_index: tuple[torch.Tensor, torch.Tensor, torch.Tensor],
        chunks: list[int],
    ) -> tuple[torch.Tensor, torch.Tensor]:
        _patient_index_full, rows_full, tasks_full = prediction_index
        list_rows = torch.split(rows_full, chunks)
        list_tasks = torch.split(tasks_full, chunks)

        output_list = []
        for chain_X in X:  # iterate through the individual chains
            # Separate the individual patients
            list_X = [ind_X for ind_X in chain_X]

            input_df_list = []
            for ind_X, ind_rows, ind_tasks in zip(list_X, list_rows, list_tasks):
                temp_id = str(uuid.uuid4())
                # Extract the parameters and time values
                params = ind_X.index_select(0, ind_rows).cpu().detach().numpy()
                # Extract the task order
                task_index = ind_tasks.cpu().detach().numpy()
                # Format the data inputs
                # This step is where the order of parameters is implicit
                input_df_temp = pd.DataFrame(
                    data=params, columns=self.parameter_names + ["time"]
                )
                # The passed params include the _global_ time steps
                # Filter the time steps that we actually want for this patient
                input_df_temp = input_df_temp.iloc[ind_rows.cpu().numpy()]
                # Add the task index as a temporary column
                input_df_temp["task_index"] = task_index
                # Deduce protocol arm and output name from task index
                input_df_temp["protocol_arm"] = input_df_temp["task_index"].apply(
                    lambda t: self.task_idx_to_protocol[t]
                )
                input_df_temp["output_name"] = input_df_temp["task_index"].apply(
                    lambda t: self.output_names[self.task_idx_to_output_idx[t]]
                )
                # Remove the unnecessary task index column
                input_df_temp = input_df_temp.drop(columns=["task_index"])
                input_df_temp["id"] = temp_id
                # Add the protocol overrides
                if self.nb_protocol_overrides > 0:
                    input_df_temp = input_df_temp.merge(
                        self.protocol_design, how="left", on=["protocol_arm"]
                    )
                # Add the initial conditions
                input_df_temp = input_df_temp.merge(self.init_cond_df, how="cross")
                input_df_list.append(input_df_temp)

            full_input = pd.concat(input_df_list)
            # Simulate the ODE model
            output_df = self.ode_model.simulate_model(full_input)
            # Convert back to tensor
            out_tensor = torch.as_tensor(
                output_df["predicted_value"].values, device=device
            )
            output_list.append(out_tensor)
        out_full = torch.stack(output_list, dim=0).to(device)
        out_var = torch.zeros_like(out_full)
        return out_full, out_var


class StructuralAnalytical(StructuralModel):
    def __init__(
        self,
        equations: Callable,
        variable_names: list[str],
        protocol_design: Optional[pd.DataFrame] = None,
    ):
        # Extracts the ORDERED list of arguments from the "equations" function signature
        # This will be used to map the columns of the input tensor to their corresponding positional arguments
        # when invoking "equations"
        function_arguments = list(
            # 1. ".__code__.co_varnames" is a tuple of ALL the variables (including local variables) referenced in
            # the function, STARTING with the function arguments which is the only reason this works.
            # 2. # ".__code__.co_argcount" is the number of function arguments
            # See the "inspect" doc for more details
            # https://docs.python.org/3/library/inspect.html
            equations.__code__.co_varnames[: equations.__code__.co_argcount]
        )
        self.equations = equations
        if protocol_design is None:
            protocol_design = pd.DataFrame({"protocol_arm": ["identity"]})
        protocol_arms = protocol_design["protocol_arm"].drop_duplicates().to_list()
        protocol_arm_idx = {
            arm: arm_index for arm_index, arm in enumerate(protocol_arms)
        }
        variable_idx = {
            variable_name: variable_index
            for variable_index, variable_name in enumerate(variable_names)
        }
        protocol_overrides_set = set(protocol_design.drop(columns="protocol_arm"))
        # the parameters of the "equations" function which are NOT protocol overrides and NOT time, in this order
        parameter_names_without_protocol_overrides = [
            p
            for p in function_arguments
            if p not in protocol_overrides_set and p != "t"
        ]
        # the parameters of the "equations" function which are protocol overrides, in this order
        protocol_parameters = [
            p for p in function_arguments if p in protocol_overrides_set
        ]
        self.nb_protocol_overrides = len(protocol_parameters)

        # time is special cases in "nlme.py" and always comes after the structural parameters
        input_parameters = (
            parameter_names_without_protocol_overrides + ["t"] + protocol_parameters
        )
        parameter_index = {p: index for (index, p) in enumerate(input_parameters)}
        self.input_tensor_column_index_to_function_parameter_index = [
            parameter_index[a] for a in function_arguments
        ]

        # The tensor of protocol overrides with shape nb_arms x nb_protocol_params
        # np.atleast_2d() ensures the array has at least shape (1, 0) if there are not protocol overrides
        # (e.g. in the "identity" protocol)
        if len(protocol_parameters) == 0:
            protocol_tensor = np.zeros((len(protocol_arms), 0))
        else:
            protocol_tensor = np.atleast_2d(
                np.transpose(
                    np.array([protocol_design[p] for p in protocol_parameters])
                )
            )
        # Building the tasks names and the various index maps in one go
        tasks = []
        task_idx_to_output_idx = {}
        task_idx_to_protocol = {}
        task_protocol_overrides = []
        for task_index, (arm, output) in enumerate(
            itertools.product(protocol_arms, variable_names)
        ):
            task_name = f"{output}_{arm}"
            tasks.append(task_name)
            task_idx_to_output_idx[task_index] = variable_idx[output]
            task_idx_to_protocol[task_index] = arm
            arm_idx = protocol_arm_idx[arm]
            task_protocol_overrides.append(np.atleast_1d(protocol_tensor[arm_idx, :]))
        # the protocol overrides are stored in tensor of shape (num_tasks, nb_protocol_params)
        # such that it can later be efficiently appended to the input tensor "X" in simulate()
        self.task_protocol_tensor = torch.as_tensor(
            np.array(task_protocol_overrides), device=device
        )

        super().__init__(
            parameter_names_without_protocol_overrides,
            variable_names,
            protocol_arms,
            tasks,
            task_idx_to_output_idx,
            task_idx_to_protocol,
        )

    def simulate(
        self,
        X: torch.Tensor,
        prediction_index: tuple[torch.Tensor, torch.Tensor, torch.Tensor],
        chunks: list[int],
    ) -> tuple[torch.Tensor, torch.Tensor]:
        (num_chains, nb_patients, nb_timesteps, nb_params) = X.shape
        patient_index, timestep_index, task_index = prediction_index
        protocol_overrides = self.task_protocol_tensor[task_index].reshape(
            num_chains, nb_patients, nb_timesteps, self.nb_protocol_overrides
        )
        X_with_protocol_overrides = torch.concat((X, protocol_overrides), dim=-1)

        nb_obs_per_chain = patient_index.shape[0]
        prediction_index_expanded = (
            torch.arange(num_chains).repeat_interleave(nb_obs_per_chain),
            patient_index.repeat(num_chains),
            timestep_index.repeat(num_chains),
            task_index.apply_(lambda i: self.task_idx_to_output_idx[i]).repeat(
                num_chains
            ),
        )
        # map the "columns" (in fact the last axis which corresponds to the parameters) of X_with_protocol_overrides
        # to the positions of the corresponding arguments in the signature of the "equations" function
        params = X_with_protocol_overrides[
            :, :, :, self.input_tensor_column_index_to_function_parameter_index
        ].split(1, dim=-1)
        outputs = self.equations(*params)
        y = outputs[prediction_index_expanded].view(num_chains, nb_obs_per_chain)
        pred_var = torch.zeros_like(y)
        return y, pred_var



from .nlme import NlmeModel
from .saem import PySaem
from .structural_model import StructuralGp, StructuralOdeModel, StructuralAnalytical
from .model import *
from .ode import OdeModel
from .vpop import generate_vpop_from_ranges
from .data_generation import simulate_dataset_from_omega, simulate_dataset_from_ranges
from .diagnostics import (
    check_surrogate_validity_gp,
    plot_map_estimates,
    plot_individual_map_estimates,
    plot_all_individual_map_estimates,
    plot_map_estimates_gof,
    plot_weighted_residuals,
    plot_map_vs_posterior,
)

__all__ = [
    "GP",
    "OdeModel",
    "StructuralGp",
    "StructuralOdeModel",
    "StructuralAnalytical",
    "NlmeModel",
    "PySaem",
    "simulate_dataset_from_omega",
    "simulate_dataset_from_ranges",
    "generate_vpop_from_ranges",
    "check_surrogate_validity_gp",
    "plot_map_estimates",
    "plot_individual_map_estimates",
    "plot_all_individual_map_estimates",
    "plot_map_estimates_gof",
    "plot_weighted_residuals",
    "plot_map_vs_posterior",
]



import pandas as pd
import numpy as np
from scipy.integrate import solve_ivp
import multiprocessing as mp
from typing import Any, Callable, Optional

from .utils import smoke_test


class OdeModel:
    def __init__(
        self,
        equations: Callable,
        variable_names: list[str],
        param_names: list[str],
        tol: Optional[float] = 1e-6,
        multithreaded: Optional[bool] = True,
    ):
        """OdeModel

        Create a computational model given a set of equations.

        Args:
            equations (callable): A function describing the right hand side of the ODE system
            variable_names (list[str]): The names of the outputs of the system
            param_names (list[str]): The name of the parameters of the system
        """
        self.equations = equations
        self.variable_names = variable_names
        self.nb_outputs = len(variable_names)

        self.param_names = param_names
        self.nb_parameters = len(param_names)
        # Define the name of initial conditions as `<variable>_0`
        self.initial_cond_names = [v + "_0" for v in self.variable_names]

        self.tol = tol
        if smoke_test:
            self.use_multiprocessing = False
        else:
            self.use_multiprocessing = multithreaded

    def simulate_model(
        self,
        input_data: pd.DataFrame,
    ) -> pd.DataFrame:
        """Solve the ODE model using formatted input data

        Args:
            input_data (pd.DataFrame): a DataFrame containing 'id', 'output_name', 'time', and columns for all individual parameters and initial conditions ('var_0', for var in variable_names) for each patient

        Returns:
            pd.DataFrame: a DataFrame with the same inputs and a new 'predicted_value' column
        """

        # group the data by individual to create tasks for each process
        tasks: list[Any] = []
        for _, ind_df in input_data.groupby("id"):
            for _, filtered_df in ind_df.groupby("protocol_arm"):
                params = filtered_df[self.param_names].iloc[0].values
                initial_conditions = filtered_df[self.initial_cond_names].iloc[0].values
                input_df = filtered_df[["id", "protocol_arm", "output_name", "time"]]
                indiv_task = {
                    "patient_inputs": input_df,
                    "initial_conditions": initial_conditions,
                    "params": params,
                    "equations": self.equations,
                    "output_names": self.variable_names,
                    "tol": self.tol,
                }
                tasks.append(indiv_task)
        if self.use_multiprocessing:
            with mp.Pool() as pool:
                all_solutions: list[pd.DataFrame] = pool.map(_simulate_patient, tasks)
        else:
            all_solutions: list[pd.DataFrame] = list(map(_simulate_patient, tasks))
        output_data = pd.concat(all_solutions)
        return output_data

    def run_trial(
        self,
        vpop: pd.DataFrame,
        initial_conditions: np.ndarray,
        protocol_design: Optional[pd.DataFrame],
        time_steps: np.ndarray,
    ) -> pd.DataFrame:
        """Run a trial given a vpop, protocol and solving times

        Args:
            vpop (pd.DataFrame): The patient descriptors. Should contain the following columns
            - `id`
            - `protocol_arm`
            - `output_name`
            - one column per patient descriptor
            initial_conditions (np.ndarray): one set of initial conditions (same for all patients)
            protocol_design (Optional[pd.DataFrame]): Protocol design linking `protocol_arm` to actual parameter overrides
            time_steps (np.ndarray): The requested observation times. Same for all outputs

        Returns:
            pd.DataFrame: A merged output containing the following columns
            - `id`
            - one column per patient descriptor
            - `protocol_arm`
            - `output_name`
            - `predicted_value`: the simulated value

        Notes:
            Each patient will be run on each protocol arm, and all outputs will be included
        """

        # list the requested time steps for each output (here we use same solving times for all outputs)
        time_steps_df = pd.DataFrame({"time": time_steps})
        # Assemble the initial conditions in a dataframe
        init_cond_df = pd.DataFrame(
            data=[initial_conditions], columns=self.initial_cond_names
        )
        if protocol_design is None:
            protocol_design_to_use = pd.DataFrame({"protocol_arm": "identity"})
        else:
            protocol_design_to_use = protocol_design

        # Merge the data frames together
        # Add time steps and output names for all patients
        full_input_data = vpop.merge(time_steps_df, how="cross")
        # Add initial conditions for all patients
        full_input_data = full_input_data.merge(init_cond_df, how="cross")
        # Add protocol arm info by merging the protocol design
        full_input_data = full_input_data.merge(
            protocol_design_to_use, how="left", on="protocol_arm"
        )
        # Run the model
        output = self.simulate_model(full_input_data)

        merged_df = pd.merge(
            full_input_data,
            output,
            on=["id", "output_name", "time", "protocol_arm"],
            how="left",
        )
        return merged_df


def _simulate_patient(args: dict) -> pd.DataFrame:
    """Worker function to simulate a model on a single patient

    Args:
        args (dict): describes the simulation to be performed. Requires the following
            patient_inputs (pd.DataFrame): a data frame describing the patient to be simulated. The output data frame will be identical, with an additional `predicted_value` column. The inputs expect the following columns
                `id`
                `protocol_arm`
                `output_name`
                `time`
            initial_conditions (dict[str,float]): the initial conditions for each variable
            params (dict[str,float]): the patients descriptors
            equations (Callable): system right-hand side
            output_names (list[str]): the model output names, in the same order as in the equations
            tol (float): solver tolerance

    Returns:
        list(dict): A list of model result entries
    """

    # extract args
    input_df: pd.DataFrame = args["patient_inputs"]
    ind_id: pd.Series = input_df["id"].drop_duplicates()
    if ind_id.shape[0] > 1:
        raise ValueError("More than 1 patient was provided to `simulate_patient`")

    time_steps: list[float] = input_df["time"].drop_duplicates().to_list()
    initial_conditions: np.ndarray = args["initial_conditions"]
    params: np.ndarray = args["params"]
    equations: Callable = args["equations"]
    output_names: list[str] = args["output_names"]
    tol: float = args["tol"]

    time_span = (time_steps[0], time_steps[-1])

    sol = solve_ivp(
        equations,
        time_span,
        initial_conditions,
        method="LSODA",
        t_eval=time_steps,
        rtol=tol,
        atol=tol,
        args=params,
    )
    if not sol.success:
        raise ValueError(f"ODE integration failed: {sol.message}")

    # Filter the solver output to keep only the requested time steps for each output
    simulation_outputs_df = pd.DataFrame(data=sol.y.transpose(), columns=output_names)
    simulation_outputs_df["time"] = time_steps
    simulation_outputs_df = simulation_outputs_df.melt(
        id_vars=["time"],
        value_vars=list(output_names),
        var_name="output_name",
        value_name="predicted_value",
    )
    full_output = input_df.merge(
        simulation_outputs_df, how="left", on=["output_name", "time"]
    )
    return full_output



import numpy as np
import pandas as pd
import uuid

from vpop_calibration import *


def test_ode_saem(np_rng):

    def equations(t, y, k_a, k_12, k_21, k_el):
        # y[0] is A_absorption, y[1] is A_central, y[2] is A_peripheral
        A_absorption, A_central, A_peripheral = y[0], y[1], y[2]
        dA_absorption_dt = -k_a * A_absorption
        dA_central_dt = (
            k_a * A_absorption
            + k_21 * A_peripheral
            - k_12 * A_central
            - k_el * A_central
        )
        dA_peripheral_dt = k_12 * A_central - k_21 * A_peripheral

        ydot = [dA_absorption_dt, dA_central_dt, dA_peripheral_dt]
        return ydot

    variable_names = ["A0", "A1", "A2"]
    parameter_names = ["k_a", "k_12", "k_21", "k_el"]

    initial_conditions = np.array([10.0, 0.0, 0.0])

    protocol_design = pd.DataFrame(
        {"protocol_arm": ["arm-A", "arm-B"], "k_el": [0.5, 10.0]}
    )

    pk_two_compartments_model = OdeModel(equations, variable_names, parameter_names)

    # Parameter definitions
    error_model_type = "additive"

    # Create a patient data frame
    # It should contain at the very minimum one `id` per patient
    nb_patients = 3
    patients_df = pd.DataFrame({"id": [str(uuid.uuid4()) for _ in range(nb_patients)]})
    patients_df["protocol_arm"] = np_rng.binomial(1, 0.5, nb_patients)
    patients_df["protocol_arm"] = patients_df["protocol_arm"].apply(
        lambda x: "arm-A" if x == 0 else "arm-B"
    )
    patients_df["k_a"] = np_rng.lognormal(-1, 0.1, nb_patients)
    patients_df["foo"] = np_rng.lognormal(0.1, 0.1, nb_patients)

    obs_df = (
        patients_df[["id", "protocol_arm"]]
        .merge(pd.DataFrame({"output_name": variable_names}), how="cross")
        .merge(pd.DataFrame({"time": [0.0, 1.0, 2.0]}), how="cross")
    )
    obs_df["value"] = np_rng.normal()

    # Initial pop estimates
    # Parameter definitions
    init_log_MI = {"k_12": -1.0}
    init_log_PDU = {
        "k_21": {"mean": -1.0, "sd": 0.2},
    }
    error_model_type = "additive"
    init_res_var = [0.1, 0.05, 0.5]
    init_covariate_map = {
        "k_21": {"foo": {"coef": "cov_foo_k12", "value": 0.1}},
    }

    # Create a structural model
    structural_ode = StructuralOdeModel(
        pk_two_compartments_model, protocol_design, initial_conditions
    )
    # Create a NLME moedl
    nlme = NlmeModel(
        structural_ode,
        patients_df,
        init_log_MI,
        init_log_PDU,
        init_res_var,
        init_covariate_map,
        error_model_type,
    )
    # Create an optimizer: here we use SAEM
    optimizer = PySaem(
        nlme,
        obs_df,
    )

    optimizer.run()



import torch
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from scipy.optimize import minimize
from tqdm.notebook import tqdm
from typing import Union, Optional, Callable
from pandas import DataFrame
import numpy as np
from IPython.display import display, DisplayHandle

from .utils import smoke_test, device
from .nlme import NlmeModel
from .structural_model import StructuralGp


# Main SAEM Algorithm Class
class PySaem:
    def __init__(
        self,
        model: NlmeModel,
        observations_df: DataFrame,
        # MCMC parameters for the E-step
        mcmc_first_burn_in: int = 5,
        mcmc_nb_transitions: int = 1,
        nb_phase1_iterations: int = 100,
        nb_phase2_iterations: Union[int, None] = None,
        convergence_threshold: float = 1e-4,
        patience: int = 5,
        learning_rate_power: float = 0.8,
        annealing_factor: float = 0.95,
        init_step_size: float = 0.5,  # stick to the 0.1 - 1 range
        verbose: bool = False,
        optim_max_fun: int = 500,
        live_plot: bool = True,
        plot_frames: int = 20,
        plot_columns: int = 3,
        plot_indiv_figsize: tuple[float, float] = (3.0, 1.2),
        true_log_MI: Optional[dict[str, float]] = None,
        true_log_PDU: Optional[dict[str, dict[str, float | bool]]] = None,
        true_res_var: Optional[list[float]] = None,
        true_covariates: Optional[dict[str, dict[str, dict[str, str | float]]]] = None,
    ):
        """Instantiate an SAEM optimizer for an NLME model

        Args:
            model (NlmeModel): The model to be optimized
            observations_df (DataFrame): The data set containing observations
            mcmc_first_burn_in (int, optional): Number of discarded samples in the first iteration. Defaults to 5.
            mcmc_nb_transitions (int, optional): Number of kernel transitions computed at each iteration. Defaults to 1.
            nb_phase1_iterations (int, optional): Number of iterations in the exploration phase. Defaults to 100.
            nb_phase2_iterations (Union[int, None], optional): Number of iterations in the convergence phase. Defaults to None, implying nb_phase_2 = nb_phase_1.
            convergence_threshold (float, optional): Estimated parameter relative change threshold considered for convergence. Defaults to 1e-4.
            patience (int, optional): Number of iterations of consecutive low relative change considered for early stopping of the algorithm. Defaults to 5.
            learning_rate_power (float, optional): Exponential decay exponent for the M-step learning rate (stochastic approximation). Defaults to 0.8.
            annealing_factor (float, optional): Exploration phase annealing factor for residual and parameter variance. Defaults to 0.95.
            init_step_size (float, optional): Initial MCMC step size scaling factor. Defaults to 0.5.
            optim_max_fun(int): Maximum number of function calls in the scipy.optimize (used for model intrinsic parameters calibration). Defaults to 50.
            verbose (bool): Print various info during iterations. Defaults to False.
            live_plot (bool): Print and update a plot of parameters during iterations. Defaults to True.
            plot_frames (int): Frequency at which the live plot should be updated (number of iterations). The lower the slower. Defaults to 20.
            plot_columns (int): Number of columns to display the convergence plot. Defaults to 3.
            plot_indiv_figsize (tuple[float,float]): individual figure size in the convergence plot (width, height).
        """

        self.model: NlmeModel = model
        self.model.add_observations(observations_df)
        # MCMC sampling in the E-step parameters
        self.mcmc_first_burn_in: int = mcmc_first_burn_in
        self.mcmc_nb_transitions: int = mcmc_nb_transitions
        # SAEM iteration parameters
        # phase 1 = exploratory: learning rate = 0 and simulated annealing on
        # phase 2 = smoothing: learning rate 1/phase2_iter^factor
        if smoke_test:
            self.nb_phase1_iterations = 1
            self.nb_phase2_iterations = 2
        else:
            self.nb_phase1_iterations: int = nb_phase1_iterations
            self.nb_phase2_iterations: int = (
                nb_phase2_iterations
                if nb_phase2_iterations is not None
                else nb_phase1_iterations
            )
        self.current_phase = 1

        # convergence parameters
        self.convergence_threshold: float = convergence_threshold
        self.patience: int = patience
        self.consecutive_converged_iters: int = 0

        # Numerical parameters that depend on the iterations phase
        # The learning rate for the step-size adaptation in E-step sampling
        self.step_size: float = init_step_size / np.sqrt(self.model.nb_PDU)
        self.init_step_size_adaptation: float = 0.5
        self.step_size_learning_rate_power: float = 0.5

        # The learning rate for the stochastic approximation in the M-step
        self.learning_rate_m_step: float = 1.0
        self.learning_rate_power: float = learning_rate_power
        self.annealing_factor: float = annealing_factor

        # Initialize the learning rate and step size adaptation rate
        self.learning_rate_m_step, self.step_size_adaptation = (
            self._compute_learning_rates(0)
        )

        self.verbose = verbose
        if smoke_test:
            self.optim_max_fun = 1
        else:
            self.optim_max_fun = optim_max_fun
        self.live_plot = live_plot
        self.plot_frames = plot_frames
        self.plot_columns = plot_columns
        self.plot_indiv_figsize = plot_indiv_figsize

        # Initialize the running variables for optimization
        self.current_etas_chains = self.model.eta_samples_chains
        (
            self.current_log_prob_chains,
            self.current_gaussian_params,
            self.current_pred,
        ) = self.model.log_posterior_etas(self.current_etas_chains)
        # Initialize the complete likelihood to a dummy value to avoid messing with the plot scale
        self.current_complete_likelihood = torch.Tensor([0])

        # Initialize the optimizer history
        self._init_history(
            self.model.population_betas,
            self.model.omega_pop,
            self.model.log_MI,
            self.model.residual_var,
            self.current_complete_likelihood,
        )
        self.current_iteration: int = 0

        # Initialize the values for convergence checks
        self.prev_params: dict[str, torch.Tensor] = {
            "log_MI": self.model.log_MI,
            "population_betas": self.model.population_betas,
            "population_omega": self.model.omega_pop,
            "residual_error_var": self.model.residual_var,
        }

        # Store the full design matrix, with shape (num_chains, nb_patient, nb_PDU, nb_betas)
        self.X = self.model.full_design_matrix
        self.X_chains = self.X.expand(self.model.num_chains, -1, -1, -1)
        # Precompute the batch gram matrix
        # Gram = sum(X_i.T @ X_i, for i in patients)
        self.sufficient_stat_gram_matrix = (
            torch.matmul(self.X.transpose(-1, -2), self.X).sum(dim=0).to(device)
        )

        # Initialize sufficient statistics by performing one M-step update (without updating beta or omega)
        self.sufficient_stat_cross_product, self.sufficient_stat_outer_product, _, _ = (
            self.m_step_update(self.current_gaussian_params)
        )

        # Store true NLME parameters, if provided
        self.true_log_MI = true_log_MI
        self.true_log_PDUs = true_log_PDU
        if true_covariates is not None:
            self.true_cov = {
                str(cov["coef"]): float(cov["value"])
                for item in true_covariates.values()
                for cov in item.values()
            }
        if true_res_var is not None:
            self.true_res_var = {
                self.model.outputs_names[k]: val for k, val in enumerate(true_res_var)
            }

    def m_step_update(
        self,
        gaussian_params: torch.Tensor,
        current_cross_product: Optional[torch.Tensor] = None,
        current_outer_product: Optional[torch.Tensor] = None,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """Perform the M-step update

        Args:
            gaussian_params (torch.Tensor): Current estimation of the gaussian parameters (over all chains)
            s_cross_product (torch.Tensor): Current sufficient statistics 1 - cross product
            s_outer_product (torch.Tensor): Current sufficient statistics 2 - outer product

        Returns:
            tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]: Updated value for
            - sufficient statistics: cross product
            - sufficient statistics: outer product
            - beta parameters
            - omega matrix
        """
        assert gaussian_params.shape == (
            self.model.num_chains,
            self.model.nb_patients,
            self.model.nb_PDU,
        )
        cross_product = (
            (self.X_chains.transpose(-1, -2) @ gaussian_params.unsqueeze(-1))
            .sum(dim=1)
            .mean(dim=0)
            .to(device)
        )
        if current_cross_product is None:
            new_s_cross_product = cross_product
        else:
            new_s_cross_product = self._stochastic_approximation(
                current_cross_product, cross_product
            )
        new_beta = torch.linalg.solve(
            self.sufficient_stat_gram_matrix, new_s_cross_product
        ).to(device)

        new_mu = torch.matmul(self.X, new_beta.unsqueeze(0)).squeeze(-1).to(device)
        new_residuals = gaussian_params - new_mu.unsqueeze(0)
        resid_unsq = new_residuals.unsqueeze(-1)
        outer_product_centered = (
            torch.matmul(resid_unsq, resid_unsq.transpose(-1, -2))
            .sum(dim=1)
            .mean(dim=0)
            .to(device)
        )
        if current_outer_product is None:
            new_s_outer_product = outer_product_centered
        else:
            new_s_outer_product = self._stochastic_approximation(
                current_outer_product, outer_product_centered
            )

        # Propose a new value for omega
        new_omega = new_s_outer_product / self.model.nb_patients
        new_omega = self._clamp_eigen_values(new_omega)

        return (
            new_s_cross_product,
            new_s_outer_product,
            new_beta.squeeze(-1),
            new_omega,
        )

    def _check_convergence(self, new_params: dict[str, torch.Tensor]) -> bool:
        """Checks for convergence based on the relative change in parameters."""
        all_converged = True
        for name, current_val in new_params.items():
            if current_val.shape[0] > 0:
                prev_val = self.prev_params[name]
                abs_diff = torch.abs(current_val - prev_val)
                abs_sum = torch.abs(current_val) + torch.abs(prev_val) + 1e-9
                relative_change = abs_diff / abs_sum
                if torch.any(relative_change > self.convergence_threshold):
                    all_converged = False
                    break
        return all_converged

    def _compute_learning_rates(self, iteration: int) -> tuple[float, float]:
        """
        Calculates the SAEM learning rate (alpha_k) and Metropolis Hastings step-size (gamma_k).

        Phase 1:
          alpha_k = 1 (exploration)
          gamma_k = c_0 / k^(0.5) , c0 = init_step_size_adaptation / sqrt(n_PDU)
        Phase 2:
          alpha_k = 1 / (iteration - phase1_iterations + 1) ^ exponent (the iteration index in phase 2)
          gamma_k = 0
        """
        if iteration < self.nb_phase1_iterations:
            learning_rate_m_step = 1.0
            learning_rate_e_step = self.init_step_size_adaptation / (
                np.maximum(1, iteration) ** 0.5
            )
        else:
            learning_rate_m_step = 1.0 / (
                (iteration - self.nb_phase1_iterations + 1) ** self.learning_rate_power
            )
            learning_rate_e_step = 0
        return learning_rate_m_step, learning_rate_e_step

    def _stochastic_approximation(
        self, previous: torch.Tensor, new: torch.Tensor
    ) -> torch.Tensor:
        """Perform stochastic approximation

        Args:
            previous (torch.Tensor): The current value of the tensor
            new (torch.Tensor): The target value of the tensor

        Returns:
            torch.Tensor: (1 - learning_rate) * previous + learning_rate * new
        """
        assert (
            previous.shape == new.shape
        ), f"Wrong shape in stochastic approximation: {previous.shape}, {new.shape}"

        stochastic_approx = (
            (1 - self.learning_rate_m_step) * previous + self.learning_rate_m_step * new
        ).to(device)
        return stochastic_approx

    def _simulated_annealing(
        self, current: torch.Tensor, target: torch.Tensor
    ) -> torch.Tensor:
        """Perform simulated annealing

        This function allows to constrain the reduction of certain values by a factor stored in self.annealing_factor

        Args:
            current (torch.Tensor): Current value of the tensor
            target (torch.Tensor): Target value of the tensor

        Returns:
            torch.Tensor: maximum(annealing_factor * current, target)
        """
        return torch.maximum(self.annealing_factor * current, target).to(device)

    def _clamp_eigen_values(self, omega: torch.Tensor, min_eigenvalue: float = 1e-6):
        """
        Project a matrix onto the cone of Positive Definite matrices.
        """
        # 1. Ensure symmetry (sometimes float error breaks symmetry slightly)
        omega = (0.5 * (omega + omega.T)).to(device)

        # 2. Eigen Decomposition
        L, V = torch.linalg.eigh(omega)

        # 3. Clamp eigenvalues
        L_clamped = torch.clamp(L, min=min_eigenvalue)

        # 4. Reconstruct
        matrix_spd = torch.matmul(V, torch.matmul(torch.diag(L_clamped), V.T))

        return matrix_spd.to(device)

    def _init_history(
        self,
        beta: torch.Tensor,
        omega: torch.Tensor,
        log_mi: torch.Tensor,
        res_var: torch.Tensor,
        likelihood: torch.Tensor,
    ) -> None:
        # Initialize the history
        self.history = {}
        # Add the pdus (mean, variance)
        for i, pdu in enumerate(self.model.PDU_names):
            beta_index = self.model.population_betas_names.index(pdu)
            self.history.update(
                {
                    pdu: {
                        "mu": [beta[beta_index].cpu()],
                        "sigma_sq": [omega[i, i].cpu()],
                    }
                }
            )
        # Add the covariates
        for i, cov in enumerate(self.model.covariate_coeffs_names):
            beta_index = self.model.population_betas_names.index(cov)
            self.history.update({cov: [beta[beta_index].cpu()]})
        # Add Omega
        self.history.update({"omega": [omega.cpu()]})
        # Add the model intrinsic params
        for i, mi in enumerate(self.model.MI_names):
            self.history.update({mi: [log_mi[i].cpu()]})
        # Add the residual variance
        for i, output in enumerate(self.model.outputs_names):
            self.history.update({output: [res_var[i].cpu()]})
        self.history.update({"complete_likelihood": [likelihood.cpu()]})

    def _append_history(
        self,
        beta: torch.Tensor,
        omega: torch.Tensor,
        log_mi: torch.Tensor,
        res_var: torch.Tensor,
        complete_likelihood: torch.Tensor,
    ) -> None:
        # Update the history
        for i, pdu in enumerate(self.model.PDU_names):
            beta_index = self.model.population_betas_names.index(pdu)
            self.history[pdu]["mu"].append(beta[beta_index].cpu())
            self.history[pdu]["sigma_sq"].append(omega[i, i].cpu())

        for i, cov in enumerate(self.model.covariate_coeffs_names):
            beta_index = self.model.population_betas_names.index(cov)
            self.history[cov].append(beta[beta_index].cpu())

        self.history["omega"].append(omega.cpu())

        for i, mi in enumerate(self.model.MI_names):
            self.history[mi].append(log_mi[i].cpu())

        for i, output in enumerate(self.model.outputs_names):
            self.history[output].append(res_var[i].cpu())
        self.history["complete_likelihood"].append(complete_likelihood.cpu())

    def one_iteration(self, k: int) -> bool:
        """Perform one iteration of SAEM

        Args:
            k (int): the iteration number
        """

        if self.verbose:
            print(f"Running iteration {k}")
        # If first iteration, consider burn in
        if k == 0:
            current_iter_burn_in = self.mcmc_first_burn_in
        else:
            current_iter_burn_in = 0

        self.learning_rate_m_step, self.step_size_adaptation = (
            self._compute_learning_rates(k)
        )

        # --- E-step: perform MCMC kernel transitions
        if self.verbose:
            print("  MCMC sampling")
            print(
                f"  Current MCMC parameters: step-size={self.step_size:.2f}, adaptation rate={self.step_size_adaptation:.2f}"
            )

        # Perform the initial burn-in
        for _ in range(current_iter_burn_in):
            (
                self.current_etas_chains,
                self.current_log_prob_chains,
                self.current_complete_likelihood,
                self.current_pred,
                self.current_gaussian_params,
                self.step_size,
            ) = self.model.mh_step(
                current_etas=self.current_etas_chains,
                current_log_prob=self.current_log_prob_chains,
                current_pred=self.current_pred,
                current_gaussian_params=self.current_gaussian_params,
                step_size=self.step_size,
                learning_rate=self.step_size_adaptation,
                verbose=self.verbose,
            )

        for _ in range(self.mcmc_nb_transitions):
            (
                self.current_etas_chains,
                self.current_log_prob_chains,
                self.current_complete_likelihood,
                self.current_pred,
                self.current_gaussian_params,
                self.step_size,
            ) = self.model.mh_step(
                current_etas=self.current_etas_chains,
                current_log_prob=self.current_log_prob_chains,
                current_pred=self.current_pred,
                current_gaussian_params=self.current_gaussian_params,
                step_size=self.step_size,
                learning_rate=self.step_size_adaptation,
                verbose=self.verbose,
            )
        # Update the model's eta and thetas
        self.model.update_eta_samples(self.current_etas_chains)

        # Compute the new patient parameter estimates by averaging over all chains
        new_thetas_chains = self.model.assemble_individual_parameters(
            self.model.gaussian_to_physical_params(
                self.current_gaussian_params, self.model.log_MI
            )
        )

        # --- M-Step: Update Population Means, Omega and Residual variance ---

        # 1. Update residual error variances
        sum_sq_res = self.model.sum_sq_residuals_chains(self.current_pred)
        assert sum_sq_res.shape == (
            self.model.nb_outputs,
        ), f"Unexpected residual shape: {sum_sq_res.shape}"
        target_res_var: torch.Tensor = (
            sum_sq_res / self.model.n_tot_observations_per_output
        )
        current_res_var: torch.Tensor = self.model.residual_var
        if k < self.nb_phase1_iterations:
            target_res_var = self._simulated_annealing(current_res_var, target_res_var)

        new_residual_error_var = self._stochastic_approximation(
            current_res_var, target_res_var
        )

        self.model.update_res_var(new_residual_error_var)
        # 2. Update sufficient statistics with stochastic approximation
        (
            self.sufficient_stat_cross_product,
            self.sufficient_stat_outer_product,
            new_beta,
            new_omega,
        ) = self.m_step_update(
            self.current_gaussian_params,
            self.sufficient_stat_cross_product,
            self.sufficient_stat_outer_product,
        )

        # Update beta
        self.model.update_betas(new_beta)

        # Update omega
        if k < self.nb_phase1_iterations:
            # Simulated annealing during phase 1
            new_omega_diag = torch.diag(new_omega).to(device)
            current_omega_diag = torch.diag(self.model.omega_pop).to(device)
            annealed_omega_diag = self._simulated_annealing(
                current_omega_diag, new_omega_diag
            )
            new_omega = torch.diag(annealed_omega_diag).to(device)
        self.model.update_omega(new_omega)

        # 3. Update fixed effects MIs
        if self.model.nb_MI > 0:
            # This step is notoriously under-optimized
            self.current_gaussian_params_per_patient = (
                self.current_gaussian_params.mean(dim=0)
            )
            objective_fun = self.build_mi_objective_function()
            target_log_MI_np = minimize(
                fun=objective_fun,
                x0=self.model.log_MI.cpu().squeeze().numpy(),
                method="L-BFGS-B",
                options={"maxfun": self.optim_max_fun},
            ).x
            target_log_MI = torch.from_numpy(target_log_MI_np).to(device)
            new_log_MI = self._stochastic_approximation(
                self.model.log_MI, target_log_MI
            )

            self.model.update_log_mi(new_log_MI)

        if self.verbose:
            print(
                f"  Updated MIs: {', '.join([f'{torch.exp(logMI).item():.4f}' for logMI in self.model.log_MI.detach().cpu()])}"
            )
            print(
                f"  Updated Betas: {', '.join([f'{beta:.4f}' for beta in self.model.population_betas.detach().cpu().numpy().flatten()])}"
            )
            print(
                f"  Updated Omega (diag): {', '.join([f'{val.item():.4f}' for val in torch.diag(self.model.omega_pop.detach().cpu())])}"
            )
            print(
                f"  Updated Residual Var: {', '.join([f'{res_var:.4f}' for res_var in self.model.residual_var.detach().cpu().numpy().flatten()])}"
            )

        # Convergence check
        new_params: dict[str, torch.Tensor] = {
            "log_MI": self.model.log_MI,
            "population_betas": self.model.population_betas,
            "population_omega": self.model.omega_pop,
            "residual_error_var": self.model.residual_var,
        }
        is_converged = self._check_convergence(new_params)

        # store history
        self._append_history(
            self.model.population_betas,
            self.model.omega_pop,
            self.model.log_MI,
            self.model.residual_var,
            self.current_complete_likelihood,
        )

        # update prev_params for the next iteration's convergence check
        self.prev_params = new_params

        if self.verbose:
            print("Iter done")
        return is_converged

    def build_mi_objective_function(self) -> Callable:
        def mi_objective_function(log_MI: np.ndarray):
            mi_tensor = torch.from_numpy(log_MI).to(device)
            # Assemble the patient parameters
            new_physical_params = self.model.gaussian_to_physical_params(
                self.current_gaussian_params, mi_tensor
            )
            new_thetas = self.model.assemble_individual_parameters(new_physical_params)
            predictions, _ = self.model.predict_outputs_from_theta(new_thetas)
            total_log_lik = (
                self.model.log_likelihood_observation(
                    predictions,
                )
                .cpu()
                .sum()
                .item()
            )

            return -total_log_lik

        return mi_objective_function

    def run(
        self,
    ) -> None:
        """
        This method starts the SAEM estimation by initiating some class attributes then calling the iterate method.
        returns self.population_betas, self.estimated_population_mus, self.population_omega, self.residual_error_var, self.history
        stores the current state of the estimation so that the iterations can continue later with the continue_iterating method.
        """
        if self.verbose:
            print("Starting SAEM Estimation...")
            print(
                f"Initial Population Betas: {', '.join([f'{beta.item():.2f}' for beta in self.model.population_betas.cpu()])}"
            )
            print(
                f"Initial Population MIs: {', '.join([f'{torch.exp(logMI).item():.2f}' for logMI in self.model.log_MI.cpu()])}"
            )
            print(f"Initial Omega:\n{self.model.omega_pop.cpu()}")
            print(f"Initial Residual Variance: {self.model.residual_var.cpu()}")

        print("Phase 1 (exploration):")
        (
            self.convergence_plot_handle,
            self.convergence_plot_fig,
            self.convergence_plot_axes,
        ) = self._build_convergence_plot(
            indiv_figsize=self.plot_indiv_figsize, n_cols=self.plot_columns
        )
        for k in tqdm(range(0, self.nb_phase1_iterations)):
            # Run iteration, do not check for convergence in the exploration phase
            _ = self.one_iteration(k)
            self.current_iteration = k
            if (self.live_plot) & (k % self.plot_frames == 0):
                self._update_convergence_plot()

        if self.nb_phase2_iterations > 0:
            self.current_phase = 2
            print("Phase 2 (smoothing):")
            for k in tqdm(
                range(
                    self.nb_phase1_iterations,
                    self.nb_phase1_iterations + self.nb_phase2_iterations,
                )
            ):
                # Run iteration
                is_converged = self.one_iteration(k)
                self.current_iteration = k

                if (self.live_plot) & (k % self.plot_frames == 0):
                    self._update_convergence_plot()
                # Check for convergence, and stop if criterion matched
                if is_converged:
                    self.consecutive_converged_iters += 1
                    if self.verbose:
                        print(
                            f"Convergence met. Consecutive iterations: {self.consecutive_converged_iters}/{self.patience}"
                        )
                    if self.consecutive_converged_iters >= self.patience:
                        print(
                            f"\nConvergence reached after {k + 1} iterations. Stopping early."
                        )
                        self._update_convergence_plot()
                        break
                else:
                    self.consecutive_converged_iters = 0
        self._update_convergence_plot()
        plt.close(self.convergence_plot_fig)
        self.print_estimates_console()
        return None

    def continue_iterating(self, nb_add_iters_ph1=0, nb_add_iters_ph2=0) -> None:
        """
        This method is to be used when the run method has already run and the user wants to further iterate.
        """
        if self.current_phase == 2:
            if nb_add_iters_ph1 > 0:
                print("Smoothing phase has started, cannot add phase 1 iterations.")
                nb_add_iters_ph1 = 0
        if self.current_phase == 1:
            if nb_add_iters_ph1 > 0:
                print("Continuing phase 1 (exploration):")
                for k in tqdm(range(self.nb_phase1_iterations + nb_add_iters_ph1)):
                    # Run iteration, do not check for convergence in the exploration phase
                    _ = self.one_iteration(k)

            print("Switching to Phase 2 (smoothing)")
            self.current_phase = 2

        if nb_add_iters_ph2 > 0:
            for k in tqdm(
                range(
                    self.nb_phase1_iterations
                    + self.nb_phase2_iterations
                    + nb_add_iters_ph1,
                    self.nb_phase1_iterations
                    + self.nb_phase2_iterations
                    + nb_add_iters_ph1
                    + nb_add_iters_ph2,
                )
            ):
                # Run iteration
                is_converged = self.one_iteration(k)
                # Check for convergence, and stop if criterion matched
                if is_converged:
                    self.consecutive_converged_iters += 1
                    if self.verbose:
                        print(
                            f"Convergence met. Consecutive iterations: {self.consecutive_converged_iters}/{self.patience}"
                        )
                    if self.consecutive_converged_iters >= self.patience:
                        print(
                            f"\nConvergence reached after {k + 1} iterations. Stopping early."
                        )
                        break
                else:
                    self.consecutive_converged_iters = 0
        return None

    def _build_convergence_plot(
        self,
        indiv_figsize: tuple[float, float] = (2.0, 1.2),
        n_cols: int = 3,
    ) -> tuple[DisplayHandle | None, Figure, np.ndarray]:
        """
        This method plots the evolution of the estimated parameters (MI, betas, omega, residual error variances) across iterations
        """
        history = self.history
        nb_MI: int = self.model.nb_MI
        nb_betas: int = self.model.nb_betas
        nb_omega_diag_params: int = self.model.nb_PDU
        nb_var_res_params: int = self.model.nb_outputs
        nb_plots = nb_MI + nb_betas + nb_omega_diag_params + nb_var_res_params + 1
        nb_cols = n_cols
        nb_rows = int(np.ceil(nb_plots / nb_cols))
        maxiter = self.nb_phase1_iterations + self.nb_phase2_iterations
        fig, axes = plt.subplots(
            nrows=nb_rows,
            ncols=nb_cols,
            figsize=(
                nb_cols * indiv_figsize[0],
                nb_rows * indiv_figsize[1],
            ),
            squeeze=False,
            sharex="all",
        )

        self.traces = {}
        plot_idx: int = 0
        # Plot the MI parameters
        for mi_name in self.model.MI_names:
            row, col = plot_idx // nb_cols, plot_idx % nb_cols
            ax = axes[row, col]
            ax.set_xlim(0, maxiter)
            MI_history = [h.item() for h in history[mi_name]]
            (tr,) = ax.plot(
                MI_history,
            )
            if hasattr(self, "true_log_MI"):
                if self.true_log_MI is not None:
                    ax.axhline(
                        y=self.true_log_MI[mi_name],
                        linestyle="--",
                    )
            ax.set_title(f"Model intrinsic {mi_name}")
            ax.grid(True)
            self.traces.update({mi_name: tr})
            plot_idx += 1
        # Plot the PDUs means
        for pdu in self.model.PDU_names:
            row, col = plot_idx // nb_cols, plot_idx % nb_cols
            ax = axes[row, col]
            ax.set_xlim(0, maxiter)
            beta_history = [h.item() for h in history[pdu]["mu"]]
            (tr,) = ax.plot(
                beta_history,
            )
            if hasattr(self, "true_log_PDUs"):
                if self.true_log_PDUs is not None:
                    ax.axhline(
                        y=self.true_log_PDUs[pdu]["mean"],
                        linestyle="--",
                    )
            ax.set_title(rf"{pdu}: $\mu$ (log)")
            ax.set_xlabel("")
            ax.grid(True)
            self.traces.update({pdu: {"mu": tr}})
            plot_idx += 1
        # Plot the PDUs sigma
        for pdu in self.model.PDU_names:
            row, col = plot_idx // nb_cols, plot_idx % nb_cols
            ax = axes[row, col]
            ax.set_xlim(0, maxiter)
            beta_history = [h.item() for h in history[pdu]["sigma_sq"]]
            (tr,) = ax.plot(
                beta_history,
            )
            if hasattr(self, "true_log_PDUs"):
                if self.true_log_PDUs is not None:
                    ax.axhline(
                        y=self.true_log_PDUs[pdu]["sd"],
                        linestyle="--",
                    )
            ax.set_title(rf"{pdu}: $\sigma^2$")
            ax.set_xlabel("")
            ax.grid(True)
            self.traces[pdu].update({"sigma_sq": tr})
            plot_idx += 1
        # Plot the coefficients of covariation
        for beta_name in self.model.covariate_coeffs_names:
            row, col = plot_idx // nb_cols, plot_idx % nb_cols
            ax = axes[row, col]
            ax.set_xlim(0, maxiter)
            beta_history = [h.item() for h in history[beta_name]]
            (tr,) = ax.plot(
                beta_history,
            )
            if hasattr(self, "true_cov"):
                if self.true_cov is not None:
                    ax.axhline(
                        y=self.true_cov[beta_name],
                        linestyle="--",
                    )
            ax.set_title(rf"{beta_name}")
            ax.set_xlabel("")
            ax.grid(True)
            self.traces.update({beta_name: tr})
            plot_idx += 1
        # Plot the residual variance
        for res_name in self.model.outputs_names:
            row, col = plot_idx // nb_cols, plot_idx % nb_cols
            ax = axes[row, col]
            ax.set_xlim(0, maxiter)
            var_res_history = [h.item() for h in history[res_name]]
            (tr,) = ax.plot(
                var_res_history,
            )
            if hasattr(self, "true_res_var"):
                if self.true_res_var is not None:
                    ax.axhline(
                        y=self.true_res_var[res_name],
                        linestyle="--",
                    )
            ax.set_title(rf"{res_name}: $\sigma^2$")
            ax.grid(True)
            self.traces.update({res_name: tr})
            plot_idx += 1
        # Plot the convergence indicator (total log prob)
        row, col = plot_idx // nb_cols, plot_idx % nb_cols
        ax = axes[row, col]
        ax.set_xlim(0, maxiter)
        convergence_ind = [h.item() for h in history["complete_likelihood"]]
        (tr,) = ax.plot(
            convergence_ind,
        )
        ax.set_title(rf"Convergence indicator")
        ax.grid(True)
        self.traces.update({"convergence_ind": tr})
        plot_idx += 1

        # Turn off extra subplots
        while plot_idx < nb_rows * nb_cols:
            row, col = plot_idx // nb_cols, plot_idx % nb_cols
            ax = axes[row, col]
            ax.set_visible(False)
            plot_idx += 1
        if not smoke_test:
            plt.tight_layout()
            handle = display(fig, display_id=True)
        else:
            handle = None
        return (handle, fig, axes)

    def _update_convergence_plot(self):
        history = self.history
        new_xaxis = np.arange(self.current_iteration + 2)
        # Plot the MI parameters
        for mi_name in self.model.MI_names:
            MI_history = [h.item() for h in history[mi_name]]
            self.traces[mi_name].set_data(new_xaxis, MI_history)
        # Plot the PDUs means
        for pdu in self.model.PDU_names:
            beta_history = [h.item() for h in history[pdu]["mu"]]
            self.traces[pdu]["mu"].set_data(new_xaxis, beta_history)
        # Plot the PDUs sigma
        for pdu in self.model.PDU_names:
            beta_history = [h.item() for h in history[pdu]["sigma_sq"]]
            self.traces[pdu]["sigma_sq"].set_data(new_xaxis, beta_history)
        # Plot the coefficients of covariation
        for beta_name in self.model.covariate_coeffs_names:
            beta_history = [h.item() for h in history[beta_name]]
            self.traces[beta_name].set_data(new_xaxis, beta_history)
        # Plot the residual variance
        for res_name in self.model.outputs_names:
            var_res_history = [h.item() for h in history[res_name]]
            self.traces[res_name].set_data(new_xaxis, var_res_history)
        conv_ind = [h.item() for h in history["complete_likelihood"]]
        self.traces["convergence_ind"].set_data(new_xaxis, conv_ind)
        if not smoke_test:
            for ax in self.convergence_plot_axes.flatten():
                ax.autoscale_view(scaley=True, scalex=False)
                ax.relim()
            if self.convergence_plot_handle is not None:
                self.convergence_plot_handle.update(self.convergence_plot_fig)

    def plot_convergence_history(
        self,
        indiv_figsize: tuple[float, float] = (2.0, 1.2),
        n_cols: int = 3,
    ):
        handle, fig, axes = self._build_convergence_plot(
            indiv_figsize=indiv_figsize, n_cols=n_cols
        )
        plt.close(fig)

    def print_estimates_console(self) -> None:
        print("Estimated values of population effects:\n")
        if self.model.nb_MI > 0:
            print("------")
            print("Model intrinsic parameters:")
            for i, mi in enumerate(self.model.MI_names):
                val_log = self.model.log_MI[i]
                print(f"{mi}: {torch.exp(val_log):.2f} (log: {val_log:.2f})")
        if self.model.nb_PDU > 0:
            print("------")
            print("PDU parameters:")
            for i, pdu in enumerate(self.model.PDU_names):
                beta_index = self.model.population_betas_names.index(pdu)
                mu_val = self.model.population_betas[beta_index]
                omega_val = self.model.omega_pop[i, i]
                std_dev = (torch.exp(omega_val) - 1) * torch.exp(2 * mu_val + omega_val)
                print(
                    f"{pdu}: mu: {torch.exp(mu_val): .2f} (log: {mu_val:.2f}), omega^2: {omega_val:.2e}, variance: {std_dev:.2f}"
                )
        if self.model.nb_covariates > 0:
            print("------")
            print("Covariate effect parameters:")
            for i, coef in enumerate(self.model.covariate_coeffs_names):
                beta_index = self.model.population_betas_names.index(coef)
                coef_val = self.model.population_betas[beta_index]
                print(f"{coef}: {coef_val:.2e}")
        print("------")
        print("Residual error model:")
        for i, output in enumerate(self.model.outputs_names):
            sigma = self.model.residual_var[i]
            print(f"{output}: {sigma:.2e}")



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from typing import Optional
from IPython.display import display, DisplayHandle

from ..utils import smoke_test


def plot_all_solutions(
    obs_vs_pred: pd.DataFrame, fig_scaling: tuple[float, float]
) -> None:
    """Plot the overlapped observations and model predictions for all patients, facetted by output and protocol.

    Args:
        obs_vs_pred (pd.DataFrame): Full data frame containing observations and predicitons from the model. Should contain the following columns
        - `id`
        - `output_name`
        - `protocol_arm`
        - `time`
        - `value`
        - `pred_mean`
    """
    outputs = obs_vs_pred["output_name"].unique().tolist()
    nb_outputs = len(outputs)
    protocol_arms = obs_vs_pred["protocol_arm"].unique().tolist()
    nb_protocol_arms = len(protocol_arms)
    patients = obs_vs_pred["id"].unique().tolist()

    n_cols = nb_outputs
    n_rows = nb_protocol_arms
    fig, axes = plt.subplots(
        n_rows,
        n_cols,
        figsize=(fig_scaling[0] * n_cols, fig_scaling[1] * n_rows),
        squeeze=False,
    )

    cmap = plt.get_cmap("Spectral")
    colors = cmap(np.linspace(0, 1, len(patients)))
    for output_index, output_name in enumerate(outputs):
        for protocol_index, protocol_arm in enumerate(protocol_arms):
            data_to_plot = obs_vs_pred.loc[
                (obs_vs_pred["output_name"] == output_name)
                & (obs_vs_pred["protocol_arm"] == protocol_arm)
            ]
            ax = axes[protocol_index, output_index]
            ax.set_xlabel("Time")
            for patient_num, patient_ind in enumerate(patients):
                patient_data = data_to_plot.loc[data_to_plot["id"] == patient_ind]
                time_vec = patient_data["time"].values
                sorted_indices = np.argsort(time_vec)
                sorted_times = time_vec[sorted_indices]
                obs_vec = patient_data["value"].values[sorted_indices]
                pred_vec = patient_data["pred_mean"].values[sorted_indices]
                ax.plot(
                    sorted_times,
                    obs_vec,
                    "+",
                    color=colors[patient_num],
                    linewidth=2,
                    alpha=0.6,
                )
                ax.plot(
                    sorted_times,
                    pred_vec,
                    "-",
                    color=colors[patient_num],
                    linewidth=2,
                    alpha=0.5,
                )

            title = f"{output_name} in {protocol_arm}"
            ax.set_title(title)
    if not smoke_test:
        plt.tight_layout()
        plt.show()
    plt.close(fig)


def plot_individual_solution(
    obs_vs_pred: pd.DataFrame, fig_scaling: tuple[float, float]
) -> None:
    """Plot the model prediction (and confidence interval) vs. the input data for a single patient"""
    outputs = obs_vs_pred["output_name"].unique().tolist()
    nb_outputs = len(outputs)
    protocol_arms = obs_vs_pred["protocol_arm"].unique().tolist()
    nb_protocol_arms = len(protocol_arms)
    patients = obs_vs_pred["id"].unique().tolist()
    assert len(patients) == 1
    patient_id = patients[0]
    ncols = nb_outputs
    nrows = nb_protocol_arms
    fig, axes = plt.subplots(
        nrows,
        ncols,
        figsize=(fig_scaling[0] * nrows, fig_scaling[1] * ncols),
        squeeze=False,
    )

    patient_params = obs_vs_pred.drop(
        columns=[
            "id",
            "output_name",
            "protocol_arm",
            "value",
            "pred_mean",
            "pred_low",
            "pred_high",
        ]
    ).drop_duplicates()

    for output_index, output_name in enumerate(outputs):
        for protocol_index, protocol_arm in enumerate(protocol_arms):
            data_to_plot = obs_vs_pred.loc[
                (obs_vs_pred["output_name"] == output_name)
                & (obs_vs_pred["protocol_arm"] == protocol_arm)
            ]
            time_steps = np.array(data_to_plot["time"].values)
            sorted_indices = np.argsort(time_steps)
            sorted_time_steps = time_steps[sorted_indices]
            ax = axes[protocol_index, output_index]
            ax.set_xlabel("Time")
            # Plot observations
            ax.plot(
                sorted_time_steps,
                data_to_plot["value"].values[sorted_indices],
                ".-",
                color="C0",
                linewidth=2,
                alpha=0.6,
                label=output_name,
            )

            # Plot model prediction
            ax.plot(
                sorted_time_steps,
                data_to_plot["pred_mean"].values[sorted_indices],
                "-",
                color="C3",
                linewidth=2,
                alpha=0.5,
                label="GP prediction for " + output_name + " (mean)",
            )
            # Add confidence interval
            ax.fill_between(
                sorted_time_steps,
                data_to_plot["pred_low"].values[sorted_indices],
                data_to_plot["pred_high"].values[sorted_indices],
                alpha=0.5,
                color="C3",
                label="GP prediction for " + output_name + " (CI)",
            )

            title = f"{output_name} in {protocol_arm} for patient {patient_id}"
            ax.set_title(title)

    if not smoke_test:
        plt.tight_layout()
        plt.show()
    plt.close(fig)


def plot_obs_vs_predicted(
    obs_vs_pred: pd.DataFrame,
    fig_scaling: tuple[float, float],
    logScale: Optional[list[bool]] = None,
) -> None:
    """Plots the observed vs. predicted values on the training or validation data set, or on a new data set."""

    outputs = obs_vs_pred["output_name"].unique().tolist()
    nb_outputs = len(outputs)
    protocol_arms = obs_vs_pred["protocol_arm"].unique().tolist()
    nb_protocol_arms = len(protocol_arms)
    patients = obs_vs_pred["id"].unique().tolist()

    n_cols = nb_outputs
    n_rows = nb_protocol_arms
    fig, axes = plt.subplots(
        n_rows,
        n_cols,
        figsize=(fig_scaling[0] * n_cols, fig_scaling[1] * n_rows),
        squeeze=False,
    )

    if not logScale:
        logScale = [True] * nb_outputs

    for output_index, output_name in enumerate(outputs):
        for protocol_index, protocol_arm in enumerate(protocol_arms):
            log_viz = logScale[output_index]
            ax = axes[protocol_index, output_index]
            ax.set_xlabel("Observed")
            ax.set_ylabel("Predicted")
            data_to_plot = obs_vs_pred.loc[
                (obs_vs_pred["protocol_arm"] == protocol_arm)
                & (obs_vs_pred["output_name"] == output_name)
            ]
            for ind in patients:
                patient_data = data_to_plot.loc[data_to_plot["id"] == ind]
                obs_vec = patient_data["value"]
                pred_vec = patient_data["pred_mean"]
                ax.plot(
                    obs_vec,
                    pred_vec,
                    "o",
                    linewidth=1,
                    alpha=0.6,
                )

            min_val = data_to_plot["value"].min().min()
            max_val = data_to_plot["value"].max().max()
            ax.plot(
                [min_val, max_val],
                [min_val, max_val],
                "-",
                linewidth=1,
                alpha=0.5,
                color="black",
            )
            ax.fill_between(
                [min_val, max_val],
                [min_val / 2, max_val / 2],
                [min_val * 2, max_val * 2],
                linewidth=1,
                alpha=0.25,
                color="black",
            )
            title = f"{output_name} in {protocol_arm}"  # More descriptive title
            ax.set_title(title)
            if log_viz:
                ax.set_xscale("log")
                ax.set_yscale("log")

    if not smoke_test:
        plt.tight_layout()
        plt.show()
    plt.close(fig)


class LossPlot:
    def __init__(self):
        self.fig, self.ax = plt.subplots(ncols=1, nrows=1)
        (tr1,) = plt.plot([0], [0], label="Training")
        (tr2,) = plt.plot([0], [0], label="Validation")
        self.traces = {
            "training": tr1,
            "validation": tr2,
        }
        plt.xlabel("Iteration")
        plt.ylabel("Loss")
        self.ax.legend()

        plt.title("Training and validation loss over Iterations")
        if not smoke_test:
            self.handle = display(self.fig, display_id=True)

    def update_plot(
        self,
        iterations: np.ndarray,
        train_losses: np.ndarray,
        validation_losses: np.ndarray,
    ):
        self.traces["training"].set_data(iterations, train_losses)
        if len(validation_losses) > 0:
            self.traces["validation"].set_data(iterations, validation_losses)
        if not smoke_test:
            assert self.handle is not None
            self.ax.relim()
            self.ax.autoscale_view(scalex=True, scaley=True)
            self.handle.update(self.fig)



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from sklearn.metrics import r2_score
import random as rand
from .nlme import NlmeModel
from .saem import PySaem
from .model.gp import GP
from .structural_model import StructuralGp
from .utils import smoke_test
import scipy.stats as stats


def check_surrogate_validity_gp(
    nlme_model: NlmeModel,
    scaling_indiv_plot: float = 3.0,
    scaling_2by2_plot: float = 2.0,
) -> tuple[dict, dict]:
    pdus = nlme_model.descriptors
    gp_model_struct = nlme_model.structural_model
    assert isinstance(
        gp_model_struct, StructuralGp
    ), "Posterior surrogate validity check only implemented for GP structural model."

    gp_model: GP = gp_model_struct.gp_model
    train_data = gp_model.data.full_df_raw[pdus].drop_duplicates()

    map_data = nlme_model.map_estimates_descriptors()
    patients = nlme_model.patients

    n_plots = len(pdus)
    n_cols = 3
    n_rows = int(np.ceil(n_plots / n_cols))

    _, axes1 = plt.subplots(
        n_rows,
        n_cols,
        squeeze=False,
        figsize=[scaling_indiv_plot * n_cols, scaling_indiv_plot * n_rows],
    )
    diagnostics = {}
    recommended_ranges = {}
    for k, param in enumerate(pdus):
        i, j = k // n_cols, k % n_cols
        train_samples = np.log(train_data[param])
        train_min, train_max = train_samples.min(axis=0), train_samples.max(axis=0)

        map_samples = np.log(map_data[param])
        flag_high = np.where(map_samples > train_max)[0]
        flag_low = np.where(map_samples < train_min)[0]
        recommend_low, recommend_high = train_min, train_max
        param_diagnostic = {}
        if flag_high.shape[0] > 0:
            param_diagnostic.update({"above": [patients[p] for p in flag_high]})
            recommend_high = map_samples.max()
        else:
            param_diagnostic.update({"above": None})
        if flag_low.shape[0] > 0:
            param_diagnostic.update({"below": [patients[p] for p in flag_low]})
            recommend_low = map_samples.min()
        else:
            param_diagnostic.update({"below": None})
        diagnostics.update({param: param_diagnostic})
        recommended_ranges.update(
            {
                param: {
                    "low": f"{recommend_low:.2f}",
                    "high": f"{recommend_high:.2f}",
                    "log": True,
                }
            }
        )

        ax = axes1[i, j]
        ax.hist([train_samples, map_samples], density=True)
        ax.axvline(train_min, linestyle="dashed", color="black")
        ax.axvline(train_max, linestyle="dashed", color="black")
        ax.set_title(f"{param}")

    _, axes2 = plt.subplots(
        n_plots,
        n_plots,
        squeeze=False,
        figsize=[scaling_2by2_plot * n_plots, scaling_2by2_plot * n_plots],
        sharex="col",
        sharey="row",
    )
    for k1, param1 in enumerate(pdus):
        train_samples_1 = np.log(train_data[param1])
        map_samples_1 = np.log(map_data[param1])
        for k2, param2 in enumerate(pdus):
            train_samples_2 = np.log(train_data[param2])
            map_samples_2 = np.log(map_data[param2])
            ax = axes2[k1, k2]
            if k1 != k2:
                # param 1 is the row -> y axis
                # param 2 is the column -> x axis
                ax.scatter(train_samples_2, train_samples_1, alpha=0.5, s=1.0)
                ax.scatter(map_samples_2, map_samples_1, s=5)
            if k2 == 0:
                ax.set_ylabel(param1)
            if k1 == len(pdus) - 1:
                ax.set_xlabel(param2)

    if not smoke_test:
        plt.tight_layout()
        plt.show()
    return diagnostics, recommended_ranges


def plot_map_estimates(
    nlme_model: NlmeModel, facet_width: float = 5.0, facet_height: float = 4.0
) -> None:
    observed = nlme_model.observations_df
    simulated_df = nlme_model.map_estimates_predictions()

    n_cols = nlme_model.nb_outputs
    n_rows = nlme_model.structural_model.nb_protocols
    _, axes = plt.subplots(
        n_rows,
        n_cols,
        figsize=(facet_width * n_cols, facet_height * n_rows),
        squeeze=False,
    )

    cmap = plt.get_cmap("Spectral")
    colors = cmap(np.linspace(0, 1, nlme_model.nb_patients))
    for output_index, output_name in enumerate(nlme_model.outputs_names):
        for protocol_index, protocol_arm in enumerate(
            nlme_model.structural_model.protocols
        ):
            obs_loop = observed.loc[
                (observed["output_name"] == output_name)
                & (observed["protocol_arm"] == protocol_arm)
            ]
            pred_loop = simulated_df.loc[
                (simulated_df["output_name"] == output_name)
                & (simulated_df["protocol_arm"] == protocol_arm)
            ]
            ax = axes[protocol_index, output_index]
            ax.set_xlabel("Time")
            patients_protocol = obs_loop["id"].drop_duplicates().to_list()
            for patient_ind in patients_protocol:
                patient_num = nlme_model.patients.index(patient_ind)
                patient_obs = obs_loop.loc[obs_loop["id"] == patient_ind]
                patient_pred = pred_loop.loc[pred_loop["id"] == patient_ind]
                time_vec = patient_obs["time"].values
                sorted_indices = np.argsort(time_vec)
                sorted_times = time_vec[sorted_indices]
                obs_vec = patient_obs["value"].values[sorted_indices]
                ax.plot(
                    sorted_times,
                    obs_vec,
                    "+",
                    color=colors[patient_num],
                    linewidth=2,
                    alpha=0.6,
                )
                if patient_pred.shape[0] > 0:
                    pred_vec = patient_pred["predicted_value"].values[sorted_indices]
                    ax.plot(
                        sorted_times,
                        pred_vec,
                        "-",
                        color=colors[patient_num],
                        linewidth=2,
                        alpha=0.5,
                    )

            title = f"{output_name} in {protocol_arm}"  # More descriptive title
            ax.set_title(title)

    if not smoke_test:
        plt.tight_layout()
        plt.show()


def plot_individual_map_estimates(
    nlme_model: NlmeModel,
    patient_num: int | None = None,
    facet_width: float = 5.0,
    facet_height: float = 4.0,
    verbose: bool = False,
) -> None:

    # Plot a random patient as default
    if patient_num is None:
        total_patient_num = len(nlme_model.patients)
        patient_num = rand.randrange(total_patient_num)

    # Filter datasets for the selected patient
    observed_df = nlme_model.observations_df
    simulated_df = nlme_model.map_estimates_predictions()
    patient_ind = nlme_model.patients[patient_num]
    patient_obs = observed_df.loc[(observed_df["id"] == patient_ind)]
    patient_pred = simulated_df.loc[(simulated_df["id"] == patient_ind)]

    # Print patient parameters if verbose selected
    if verbose:
        patient_params = nlme_model.map_estimates_descriptors()
        print(patient_params.loc[patient_params["id"] == patient_ind])

    # Initialize subplots
    n_cols = nlme_model.nb_outputs
    n_rows = 1
    fig, axes = plt.subplots(
        n_rows,
        n_cols,
        figsize=(facet_width * n_cols, facet_height * n_rows),
        squeeze=False,
    )
    fig.suptitle(f"Outputs for patient {patient_num}")

    # Initialize colormap according to outputs
    cmap = plt.get_cmap("brg")
    colors = cmap(np.linspace(0, 1, len(nlme_model.outputs_names)))

    for output_index, output_name in enumerate(nlme_model.outputs_names):
        # Filter dataset on current output
        patient_obs_output = patient_obs.loc[
            (patient_obs["output_name"] == output_name)
        ]
        patient_pred_output = patient_pred.loc[
            (patient_pred["output_name"] == output_name)
        ]

        # Sort dataset w.r.t time
        time_vec = patient_obs_output["time"].to_numpy()
        sorted_indices = np.argsort(time_vec)
        sorted_times = time_vec[sorted_indices]

        ax = axes[0, output_index]
        ax.set_xlabel("Time")

        # Plot observed and predicted vectors
        if patient_obs_output.shape[0] > 0:
            obs_vec = patient_obs_output["value"].values[sorted_indices]
            ax.plot(
                sorted_times,
                obs_vec,
                "+",
                color=colors[output_index],
                linewidth=2,
                alpha=0.6,
            )

        if patient_pred_output.shape[0] > 0:
            pred_vec = patient_pred_output["predicted_value"].values[sorted_indices]
            ax.plot(
                sorted_times,
                pred_vec,
                "-",
                color=colors[output_index],
                linewidth=2,
                alpha=0.5,
            )

        title = f"{output_name}"
        ax.set_title(title)
        plt.tight_layout()

    if not smoke_test:
        plt.show()

    plt.close(fig)


def plot_all_individual_map_estimates(
    nlme_model: NlmeModel,
    n_rows: int = 1,
    n_cols: int = 5,
    n_patients_to_plot: int | None = None,
    facet_width: float = 5.0,
    facet_height: float = 4.0,
    randomize: bool = False,
) -> None:

    observed_df = nlme_model.observations_df
    simulated_df = nlme_model.map_estimates_predictions()

    total_patient_num = len(nlme_model.patients)

    # Plot all patients by default
    if n_patients_to_plot is None or n_patients_to_plot > total_patient_num:
        n_patients_to_plot = total_patient_num

    print(
        f"There are {total_patient_num} patients in total. {n_patients_to_plot} will be plotted."
    )

    # Raise an error if too many patients for the grid
    if n_patients_to_plot > n_rows * n_cols:
        raise ValueError(
            f"{n_patients_to_plot} patients cannot be plotted in a {n_rows}x{n_cols} grid. Enter a n_patients_to_plot value under {n_rows*n_cols} or use a larger grid."
        )

    if randomize:
        ind_to_plot = rand.sample(range(total_patient_num), n_patients_to_plot)
    else:
        ind_to_plot = list(range(n_patients_to_plot))

    cmap = plt.get_cmap("brg")
    colors = cmap(np.linspace(0, 1, len(nlme_model.outputs_names)))

    # One plot for each output, containing all individual patients subplots for this output
    for output_index, output_name in enumerate(nlme_model.outputs_names):
        fig, axes = plt.subplots(
            n_rows,
            n_cols,
            figsize=(facet_width * n_cols, facet_height * n_rows),
            squeeze=False,
        )
        fig.suptitle(f"Output: {output_name}")

        patient_obs_output = observed_df.loc[
            (observed_df["output_name"] == output_name)
        ]
        patient_pred_output = simulated_df.loc[
            (simulated_df["output_name"] == output_name)
        ]

        for k in range(0, n_patients_to_plot):
            # Change indexing from 1d to 2d
            i = k // n_cols
            j = k % n_cols
            ax = axes[i, j]
            ax.set_xlabel("Time")

            # Filter dataset for current patient
            patient_ind = nlme_model.patients[ind_to_plot[k]]
            patient_obs = patient_obs_output.loc[
                patient_obs_output["id"] == patient_ind
            ]
            patient_pred = patient_pred_output.loc[
                patient_pred_output["id"] == patient_ind
            ]

            time_vec = patient_obs["time"].to_numpy()
            sorted_indices = np.argsort(time_vec)
            sorted_times = time_vec[sorted_indices]

            if patient_obs.shape[0] > 0:
                obs_vec = patient_obs["value"].values[sorted_indices]
                ax.plot(
                    sorted_times,
                    obs_vec,
                    "+",
                    color=colors[output_index],
                    linewidth=2,
                    alpha=0.6,
                )
            if patient_pred.shape[0] > 0:
                pred_vec = patient_pred["predicted_value"].values[sorted_indices]
                ax.plot(
                    sorted_times,
                    pred_vec,
                    "-",
                    color=colors[output_index],
                    linewidth=2,
                    alpha=0.5,
                )

            title = f"patient {ind_to_plot[k]}"
            ax.set_title(title)
            plt.tight_layout()
        if not smoke_test:
            plt.show()

        plt.close(fig)


def plot_map_estimates_gof(
    nlme_model: NlmeModel,
    facet_width: float = 8.0,
    facet_height: float = 8.0,
    tolerance_ribbon: str = "mean",
    tolerance_pct: int = 50,
) -> None:

    observed_df = nlme_model.observations_df
    simulated_df = nlme_model.map_estimates_predictions()
    sim_vs_obs_df = simulated_df.merge(observed_df, on=["time", "id", "output_name"])

    unique_outputs = sim_vs_obs_df["output_name"].unique()
    num_plots = len(unique_outputs)
    fig, axes = plt.subplots(
        1, num_plots, figsize=(facet_width * num_plots, facet_height), squeeze=False
    )

    fig.suptitle(f"Observed vs. simulated plot")

    for output_index, output_name in enumerate(nlme_model.outputs_names):

        ax = axes[0, output_index]
        gof_df = sim_vs_obs_df.loc[(sim_vs_obs_df["output_name"] == output_name)]

        # Compute R² and RMSE
        r2 = r2_score(gof_df["value"], gof_df["predicted_value"])
        rmse = np.sqrt(np.mean((gof_df["value"] - gof_df["predicted_value"]) ** 2))
        metrics_text = f"$R^2 = {r2:.3f}$\n$RMSE= {rmse:.3f}$"

        # Plot (obs,pred) points
        ax.scatter(
            x=gof_df["value"],
            y=gof_df["predicted_value"],
            alpha=0.7,
            s=50,
            edgecolors="w",
        )

        # Plot tolerance interval
        all_vals = gof_df[["value", "predicted_value"]]
        min_val = all_vals.min().min()
        max_val = all_vals.max().max()

        margin = (max_val - min_val) * 0.05
        range_val = [min_val - margin, max_val + margin]

        match tolerance_ribbon:
            case "relative":
                tol = [i * tolerance_pct / 100 for i in range_val]
            case "median":
                tol = (
                    all_vals["value"].median()
                    * tolerance_pct
                    / 100
                    * np.ones_like(range_val)
                )
            case "mean":
                tol = (
                    all_vals["value"].mean()
                    * tolerance_pct
                    / 100
                    * np.ones_like(range_val)
                )
            case _:
                tol = np.zeros_like(range_val)
        lower_bound = [val - tolerance for val, tolerance in zip(range_val, tol)]
        upper_bound = [val + tolerance for val, tolerance in zip(range_val, tol)]

        ax.plot(range_val, range_val, color="red", linestyle="-", linewidth=1.5)
        ax.fill_between(
            range_val,
            lower_bound,
            upper_bound,
            color="grey",
            linestyle="--",
            linewidth=1.5,
            alpha=0.15,
            label=f"CI: {tolerance_pct} % {tolerance_ribbon}",
        )

        ax.set_xlim(range_val)
        ax.set_ylim(range_val)
        ax.grid(True, linestyle=":", alpha=0.6)

        ax.set_xlabel("observed", fontsize=12)
        ax.set_ylabel("simulated", fontsize=12)

        ax.text(
            0.95,
            0.05,
            metrics_text,
            transform=ax.transAxes,
            verticalalignment="bottom",
            horizontalalignment="right",
            bbox=dict(
                boxstyle="round", facecolor="white", alpha=0.7, edgecolor="lightgray"
            ),
            fontsize=11,
        )

        title = f"Output: {output_name}"
        ax.legend()
        ax.set_title(title)
        plt.tight_layout()

    if not smoke_test:
        plt.show()

    plt.close(fig)


def plot_weighted_residuals(
    nlme_model: NlmeModel,
    res_type: str,
    num_samples: int = 100,
    facet_width: int = 10,
    facet_height: int = 10,
) -> None:

    match res_type:
        case "pwres":
            if smoke_test:
                num_samples = 2
            wres_results = nlme_model.compute_pwres(num_samples)
            compare_to_pop_pred = True
        case "iwres":
            wres_results = nlme_model.compute_iwres()
            compare_to_pop_pred = False
        case "npde":
            if smoke_test:
                num_samples = 2
            wres_results = nlme_model.compute_npde(num_samples)
            compare_to_pop_pred = True
        case _:
            raise ValueError(f"Not implemented residual type: {res_type}")

    all_wres = np.concatenate([p[res_type] for p in wres_results.values()]).flatten()
    all_times = np.concatenate([p["time"] for p in wres_results.values()]).flatten()

    fig, ax = plt.subplots(2, 2, figsize=(facet_width, facet_height))

    ## Histogram plot
    ax[0, 0].hist(
        all_wres, bins=30, density=True, alpha=0.6, color="skyblue", edgecolor="black"
    )
    mu, std = 0, 1
    x = np.linspace(min(all_wres), max(all_wres), 100)
    p = stats.norm.pdf(x, mu, std)
    ax[0, 0].plot(x, p, "r", linewidth=2, label=r"$\mathcal{N}(0,1)$")
    ax[0, 0].set_title(f"{res_type.upper()} distribution")
    ax[0, 0].set_xlabel("Residual values")
    ax[0, 0].set_ylabel("Density")
    ax[0, 0].legend()

    ## Q-Q plot
    stats.probplot(all_wres, dist="norm", plot=ax[1, 0])
    ax[1, 0].set_title(f"{res_type.upper()} Q-Q Plot")

    ## Plot vs. time
    ax[0, 1].grid(True, linestyle="--", alpha=0.6, which="both")
    ax[0, 1].set_facecolor("#fdfdfd")
    ax[0, 1].scatter(
        all_times,
        all_wres,
        alpha=0.5,
        color="#2c3e50",
        edgecolors="white",
        s=45,
        zorder=3,
    )
    ax[0, 1].axhline(y=0, color="black", linestyle="-", linewidth=1.5, zorder=4)
    ax[0, 1].axhline(
        y=1.96,
        color="#e74c3c",
        linestyle="--",
        linewidth=1.3,
        label=r"95% CI Limit ($\pm 1.96$)",
    )
    ax[0, 1].axhline(y=-1.96, color="#e74c3c", linestyle="--", linewidth=1.3)
    ax[0, 1].set_xlabel("Time", fontsize=12)
    ax[0, 1].set_ylabel("Weighted Residual (Standard Deviations)", fontsize=12)
    ax[0, 1].set_ylim(-1.1 * max(abs(all_wres)), 1.1 * max(abs(all_wres)))
    ax[0, 1].legend(loc="upper right", frameon=True, facecolor="white", framealpha=0.9)
    ax[0, 1].set_title(f"{res_type.upper()} vs. Time")

    ## Plot vs. predictions

    if compare_to_pop_pred:
        all_predictions = nlme_model.population_predictions()
    else:
        all_predictions = nlme_model.map_estimates_predictions()

    # Transform WRES dict into a dataframe
    rows = []
    for patient_id, content in wres_results.items():
        for wres, time in zip(content[res_type], content["time"]):
            rows.append({"id": patient_id, res_type: wres, "time": time})

    wres_df = pd.DataFrame(rows)

    # Merge WRES with predictions, matching patientID and time
    vs_pred_plot_df = pd.merge(
        wres_df, all_predictions[["id", "time", "predicted_value"]], on=["id", "time"]
    )

    wres_to_plot = vs_pred_plot_df[res_type]
    pred_to_plot = vs_pred_plot_df["predicted_value"]
    ax[1, 1].set_facecolor("#fdfdfd")
    ax[1, 1].scatter(
        pred_to_plot,
        wres_to_plot,
        alpha=0.5,
        color="#2c3e50",
        edgecolors="white",
        s=45,
        zorder=3,
    )
    ax[1, 1].axhline(y=0, color="black", linestyle="-", linewidth=1.5, zorder=4)
    ax[1, 1].axhline(
        y=1.96,
        color="#e74c3c",
        linestyle="--",
        linewidth=1.3,
        label=r"95% CI Limit ($\pm 1.96$)",
    )
    ax[1, 1].axhline(y=-1.96, color="#e74c3c", linestyle="--", linewidth=1.3)

    if compare_to_pop_pred:
        ax[1, 1].set_xlabel("Population Predictions")
    else:
        ax[1, 1].set_xlabel("Individual Predictions")

    ax[1, 1].set_ylabel("Weighted Residual (Standard Deviations)")
    ax[1, 1].set_ylim(-1.1 * max(abs(wres_to_plot)), 1.1 * max(abs(wres_to_plot)))
    ax[1, 1].legend(loc="upper right", frameon=True, facecolor="white", framealpha=0.9)
    ax[1, 1].set_title(f"{res_type.upper()} vs. Predictions")

    plt.tight_layout()

    if not smoke_test:
        plt.show()

    plt.close(fig)


def plot_map_vs_posterior(
    nlme_model: NlmeModel,
    nb_samples: int = 1000,
    n_patients_to_plot: int = 3,
):
    """
    Plot pdf of the conditional distribution along with MAP estimate, one plot per patient PDU.

    Args:

        nlme_model (NlmeModel)
        nb_samples (int): number of conditional distribution samples, used to apprixmate the pdf
        n_patients_to_plot (int): number of patients to plot, randomly selected within the population
    """

    # Sample new etas, compute physical
    sample_etas = nlme_model.sample_conditional_distribution(nb_samples)
    sample_gaussian = nlme_model.etas_to_gaussian_params(sample_etas)
    sample_physical = nlme_model.gaussian_to_physical_params(
        sample_gaussian, nlme_model.log_MI
    )

    total_patient_num = len(nlme_model.patients)
    ind_to_plot = rand.sample(range(total_patient_num), n_patients_to_plot)

    # Get EBE estimates for descriptors
    ebe_theta = nlme_model.ebe_estimates

    nb_samples, nb_patients, nb_params = sample_physical.shape

    nb_PDU = nlme_model.nb_PDU
    for k in range(n_patients_to_plot):
        patient_samples = sample_physical[:, ind_to_plot[k], :].detach().cpu().numpy()

        # Adapt rows to columns
        n_cols = 3
        n_rows = (nb_PDU + n_cols - 1) // n_cols
        _, axes = plt.subplots(n_rows, n_cols, figsize=(15, 4 * n_rows))
        axes = np.atleast_1d(axes).flatten()

        # Plot distribution and MAP for each PDU
        for i in range(len(axes)):
            ax = axes[i]
            if i < nb_PDU:
                param_data = patient_samples[:, i]

                if np.unique(param_data).size > 1:
                    kde = stats.gaussian_kde(param_data)
                    x_range = np.linspace(param_data.min(), param_data.max(), 200)
                    ax.plot(
                        x_range, kde(x_range), color="blue", lw=1.5, label="PDF (KDE)"
                    )

                map_val = ebe_theta[0][ind_to_plot[k]][i]

                ax.axvline(
                    map_val,
                    color="red",
                    linewidth=1.5,
                    linestyle="dashed",
                    label=f"MAP estimate: {map_val:.2f}",
                )

                ax.axvline(
                    param_data.mean(),
                    color="blue",
                    linewidth=1.5,
                    linestyle="dashed",
                    label=f"Conditional mean: {param_data.mean():.2f}",
                )

                ci_low, ci_high = np.percentile(param_data, [2.5, 97.5])
                ax.axvspan(
                    ci_low,
                    ci_high,
                    color="gray",
                    alpha=0.2,
                    label=f"95% CI: [{ci_low:.2f}, {ci_high:.2f}]",
                )

                ax.set_title(f"Patient {ind_to_plot[k]} - {nlme_model.PDU_names[i]}")
                ax.legend(fontsize="small")

            else:
                # Hide plot if empty
                ax.set_visible(False)

        plt.tight_layout()
        plt.show()
